/*
 *
 * This is part of the ANL/MCS/FL CAVEav distribution.
 * written by Ivan R. Judson
 *
 */

#include <netCAVcam.h>

#include <stdio.h>

#define CONFDATA "option add Vic.bandwidth %d startupFile \
\noption add Vic.mtu 4096 startupFile \
\noption add Vic.framerate %d startupFile \
\noption add Vic.defaultDevice shmem \
\noption add Vic.transmitOnStartup True\
\n\nproc user_hook {} { \
\n\tglobal inputDeviceList \
\n\tforeach v $inputDeviceList { \
\n\t\tif { [$v nickname] == \"shmem\" } { \
\n\t\t\t$v id %d \n\t\t} \n\t}\n} "


netCAVcam::netCAVcam(char *hostname, int port,
                     int fr, int bw, int ttl,
                     float ex, float ey, float ez, 
                     float cx, float cy, float cz, 
                     float ux, float uy, float uz) : CAVcam(ex, ey, ez,
                                                            cx, cy, cz,
                                                            ux, uy, uz)
{
  _shmkey = lrand48();

  _tmpbuf = new unsigned char[640 * 480 * 4];

  _host = hostname;
  _port = port;

  _fn = tempnam("/tmp/","CAM");

  _frameRate = fr;
  _bandwidth = bw;
  _ttl = ttl;

  _shmem = new FLSharedMemory(_shmkey, sizeof(CAVImagePool), 
                               0666);

  _imagepool = (CAVImagePool *)_shmem->GetAddress();

  _imagepool->_nImages = 10;
  _imagepool->_totalPoolSize = sizeof(CAVImagePool);
  _imagepool->_nextImage = _imagepool->_currentSerial = -1;
  _imagepool->Resize(320, 240, 4);
}

netCAVcam::~netCAVcam()
{
  delete _shmem;

  unlink(_fn);
  kill(_pid, SIGKILL);
}

void netCAVcam::Init(void)
{
  FILE *fp;
  char cmdpath[256];
  char cmdargs[2][128];
  char vicconf[1024];

#ifdef DEBUG
  fprintf(stderr, "Entering netCAVcam Init function.\n");
#endif

sprintf(cmdpath,"%s",  VICPATH);

#ifdef DEBUG
  fprintf(stderr, "Opening vic configuration file...");
#endif

  if((fp = fopen(_fn, "w")) != NULL) 
    {
      sprintf(vicconf, CONFDATA, _bandwidth, _frameRate, _shmem->GetID());
      fprintf(fp, "%s\n", vicconf);
      fclose(fp);
    } else {
      perror("Can't open file");
      CAVEExit();
    }

#ifdef DEBUG
  fprintf(stderr, "done.\nBuilding vic command line...");
#endif

  sprintf(cmdargs[0], "-u %s", _fn);
  sprintf(cmdargs[1], "%s/%d/%s/%d", _host, _port, "h261", _ttl);

#ifdef DEBUG
  fprintf(stderr, "done.\nForking vic process...");
#endif
  _pid = fork();

  if(!_pid)
    {
#ifdef DEBUG
      fprintf(stderr, "done.\nExecing vic...");
      fprintf(stderr, "%s %s %s\n", cmdpath, cmdargs[0], cmdargs[1]);
#endif
      if(execl(cmdpath, cmdpath, cmdargs[0], cmdargs[1], NULL) == -1) 
	{
	  perror("exec failed");
	  CAVEExit();
	}
#ifdef DEBUG
      fprintf(stderr, "done.\n");
#endif
    }

  _captureFrames = 1;

#ifdef DEBUG
  fprintf(stderr, "Successfully exiting netCAVcam Init function.\n");
#endif
}

void netCAVcam::Process(void)
{
  GLint buff_id[1];
  CAVImageDescriptor *image;
  int width, height;
  int nextIndex;
  GLint vpp[4];
  int x_adj, y_adj;

  if( _imagepool->_clientActive &&
      CAVEEye == CAVE_LEFT_EYE && CAVEMasterDisplay() &&
      !(_fc % _modFrames) && _captureFrames ) 
    {
      // Get the right place in the ring buffers to put the newest image
      nextIndex = (_imagepool->_nextImage + 1) % _imagepool->_nImages;
      image = &(_imagepool->_images[nextIndex]);

      // Grab values once
      width = image->GetWidth();
      height = image->GetHeight();

      // Get window info
      glGetIntegerv(GL_VIEWPORT, vpp);
      if(vpp[2] > width)
	x_adj = (vpp[2] - width) / 2;
      else 
	x_adj = 0;
      if(vpp[3] > height)
	y_adj = (vpp[3] - height) / 2;
      else 
	y_adj = 0;

#ifdef DEBUG
      fprintf(stderr, "Using buffer %d\n", nextIndex);
      fprintf(stderr, "Width < %d > Height < %d >\n", width, height);
      fprintf(stderr, "Shifting < %d %d > for viewport adjustment\n", 
	      x_adj, y_adj);
      fprintf(stderr, "Camera Position: < %f %f %f >\n", _eyex, _eyey, _eyez);
      fprintf(stderr, "Camera Forward: < %f %f %f >\n", _eyex + _centerx, 
	      _eyey + _centery, _eyez + _centerz);
      fprintf(stderr, "Camera Up: < %f %f %f >\n", _upx, _upy, _upz);
#endif

      glPushMatrix();
      {
	glMatrixMode(GL_MODELVIEW_MATRIX);
	glLoadIdentity();

	// Set Camera
	gluLookAt(_eyex, _eyey, _eyez, 
		  _eyex + _centerx, _eyey + _centery, _eyez + _centerz, 
		  _eyex + _upx, _eyey + _upy, _eyez + _upz);
      
	// Get which buffer we'll be reading from
	glGetIntegerv(GL_DRAW_BUFFER, buff_id);
	glReadBuffer(buff_id[0]);

	// Draw the scene
	CallDraw();

	// Wait for all graphics to get drawn
	glFinish();

	// Read the pixels from the screen
	glReadPixels(0 + x_adj, 0 + y_adj, width, height, 
		     GL_ABGR_EXT, GL_UNSIGNED_BYTE, _tmpbuf);
      }
      glPopMatrix();

      // turn the pixels around
      FlipFrame(_tmpbuf, nextIndex);
    
      // Update ring buffers
      _imagepool->_nextImage = nextIndex;
      _imagepool->_currentSerial++;
#ifdef DEBUG
      fprintf(stderr, "serial < %d >\n", _imagepool->_currentSerial);
#endif
      Update();
    }
}


