/*
 *
 * This is part of the ANL/MCS/FL CAVEav distribution.
 * written by Ivan R. Judson
 *
 */

#include <shmCAVcam.h>

shmCAVcam::shmCAVcam(int _shkey, 
		       int width, int height, int depth, 
		       float ex, float ey, float ez, 
		       float cx, float cy, float cz, 
		       float ux, float uy, float uz) : CAVcam(ex, ey, ez,
							       cx, cy, cz,
							       ux, uy, uz)
{
  _shmkey = _shkey;
  if(_shmkey == 0)
    _shmkey = lrand48();

  _fb_width = width; _fb_height = height; _fb_depth = depth;
  _fb_size = _fb_width * _fb_height;
  _tmpbuf = new unsigned char[_fb_width * _fb_height * _fb_depth];

  _shmem = new FLSharedMemory(_shmkey, 
			       sizeof(CAVImagePool), 
			       0666);
  _imagepool = (CAVImagePool *)_shmem->GetAddress();

  _imagepool->_nImages = 10;
  _imagepool->_totalPoolSize = sizeof(CAVImagePool);
  _imagepool->_nextImage = -1;
  _imagepool->_currentSerial = -1;
}

shmCAVcam::~shmCAVcam()
{
  delete _shmem;
}

void shmCAVcam::Process(void)
{
  GLint buff_id[1];
  CAVImageDescriptor *image;
  int width, height;
  int nextIndex;
  GLint vpp[4];
  int x_adj, y_adj;

  if( _imagepool->_clientActive &&
      CAVEEye == CAVE_LEFT_EYE && CAVEMasterDisplay() &&
      !(_fc % _modFrames) && _captureFrames ) 
    {
      // Get the right place in the ring buffers to put the newest image
      nextIndex = (_imagepool->_nextImage + 1) % _imagepool->_nImages;
      image = &(_imagepool->_images[nextIndex]);

      // Grab values once
      width = image->GetWidth();
      height = image->GetHeight();

      // Get window info
      glGetIntegerv(GL_VIEWPORT, vpp);
      if(vpp[2] > width)
	x_adj = (vpp[2] - width) / 2;
      else 
	x_adj = 0;
      if(vpp[3] > height)
	y_adj = (vpp[3] - height) / 2;
      else 
	y_adj = 0;

#ifdef DEBUG
      fprintf(stderr, "Using buffer %d\n", nextIndex);
      fprintf(stderr, "Width < %d > Height < %d >\n", width, height);
      fprintf(stderr, "Shifting < %d %d > for viewport adjustment\n", 
	      x_adj, y_adj);
      fprintf(stderr, "Camera Position: < %f %f %f >\n", _eyex, _eyey, _eyez);
      fprintf(stderr, "Camera Forward: < %f %f %f >\n", _eyex + _centerx, 
	      _eyey + _centery, _eyez + _centerz);
      fprintf(stderr, "Camera Up: < %f %f %f >\n", _upx, _upy, _upz);
#endif

      glPushMatrix();
      {
	glMatrixMode(GL_MODELVIEW_MATRIX);
	glLoadIdentity();

	// Set Camera
	gluLookAt(_eyex, _eyey, _eyez, 
		  _eyex + _centerx, _eyey + _centery, _eyez + _centerz, 
		  _eyex + _upx, _eyey + _upy, _eyez + _upz);
      
	// Get which buffer we'll be reading from
	glGetIntegerv(GL_DRAW_BUFFER, buff_id);
	glReadBuffer(buff_id[0]);

	// Draw the scene
	CallDraw();

	// Wait for all graphics to get drawn
	glFinish();

	// Read the pixels from the screen
	glReadPixels(0 + x_adj, 0 + y_adj, width, height, 
		     GL_ABGR_EXT, GL_UNSIGNED_BYTE, _tmpbuf);
      }
      glPopMatrix();

      // turn the pixels around
      FlipFrame(_tmpbuf, nextIndex);
    
      // Update ring buffers
      _imagepool->_nextImage = nextIndex;
      _imagepool->_currentSerial++;
    
      Update();
    }
}



